@@ -1,19 +1,71 @@
 import os, math, urllib, re
 
-from dulwich.errors import HangupException
+from dulwich.errors import HangupException, GitProtocolError
 from dulwich.index import commit_tree
 from dulwich.objects import Blob, Commit, Tag, Tree, parse_timezone
 from dulwich.pack import create_delta, apply_delta
 from dulwich.repo import Repo
 from dulwich import client
 
-from hgext import bookmarks
+try:
+    from mercurial import bookmarks
+    bookmarks.update
+    from mercurial import commands
+except ImportError:
+    from hgext import bookmarks
+try:
+    from mercurial.error import RepoError
+except ImportError:
+    from mercurial.repo import RepoError
+
 from mercurial.i18n import _
 from mercurial.node import hex, bin, nullid
 from mercurial import context, util as hgutil
 from mercurial import error
 
+import _ssh
 import util
+from overlay import overlayrepo
+
+class GitProgress(object):
+    """convert git server progress strings into mercurial progress"""
+    def __init__(self, ui):
+        self.ui = ui
+
+        self.lasttopic = None
+        self.msgbuf = ''
+
+    def progress(self, msg):
+        # 'Counting objects: 33640, done.\n'
+        # 'Compressing objects:   0% (1/9955)   \r
+        msgs = re.split('[\r\n]', self.msgbuf + msg)
+        self.msgbuf = msgs.pop()
+
+        for msg in msgs:
+            td = msg.split(':', 1)
+            data = td.pop()
+            if not td:
+                self.flush(data)
+                continue
+            topic = td[0]
+
+            m = re.search('\((\d+)/(\d+)\)', data)
+            if m:
+                if self.lasttopic and self.lasttopic != topic:
+                    self.flush()
+                self.lasttopic = topic
+
+                pos, total = map(int, m.group(1, 2))
+                util.progress(self.ui, topic, pos, total=total)
+            else:
+                self.flush(msg)
+
+    def flush(self, msg=None):
+        if self.lasttopic:
+            util.progress(self.ui, self.lasttopic, None)
+        self.lasttopic = None
+        if msg:
+            self.ui.note(msg + '\n')
 
 class GitHandler(object):
     mapfile = 'git-mapfile'
@@ -23,13 +75,15 @@ class GitHandler(object):
         self.repo = dest_repo
         self.ui = ui
 
-        if ui.config('git', 'intree'):
+        if ui.configbool('git', 'intree'):
             self.gitdir = self.repo.wjoin('.git')
         else:
             self.gitdir = self.repo.join('git')
 
         self.paths = ui.configitems('paths')
 
+        self.branch_bookmark_suffix = ui.config('git', 'branch_bookmark_suffix')
+
         self.load_map()
         self.load_tags()
 
@@ -66,8 +120,10 @@ class GitHandler(object):
         file = self.repo.opener(self.mapfile, 'w+', atomictemp=True)
         for hgsha, gitsha in sorted(self._map_hg.iteritems()):
             file.write("%s %s\n" % (gitsha, hgsha))
-        file.rename()
-
+        # If this complains that NoneType is not callable, then
+        # atomictempfile no longer has either of rename (pre-1.9) or
+        # close (post-1.9)
+        getattr(file, 'rename', getattr(file, 'close', None))()
 
     def load_tags(self):
         self.tags = {}
@@ -81,7 +137,10 @@ class GitHandler(object):
         for name, sha in sorted(self.tags.iteritems()):
             if not self.repo.tagtype(name) == 'global':
                 file.write("%s %s\n" % (sha, name))
-        file.rename()
+        # If this complains that NoneType is not callable, then
+        # atomictempfile no longer has either of rename (pre-1.9) or
+        # close (post-1.9)
+        getattr(file, 'rename', getattr(file, 'close', None))()
 
     ## END FILE LOAD AND SAVE METHODS
 
@@ -89,6 +148,7 @@ class GitHandler(object):
 
     def import_commits(self, remote_name):
         self.import_git_objects(remote_name)
+        self.update_hg_bookmarks(self.git.get_refs())
         self.save_map()
 
     def fetch(self, remote, heads):
@@ -96,6 +156,7 @@ class GitHandler(object):
         refs = self.fetch_pack(remote, heads)
         remote_name = self.remote_name(remote)
 
+        oldrefs = self.git.get_refs()
         if refs:
             self.import_git_objects(remote_name, refs)
             self.import_tags(refs)
@@ -105,11 +166,28 @@ class GitHandler(object):
             elif not self.paths:
                 # intial cloning
                 self.update_remote_branches('default', refs)
-        else:
-            self.ui.status(_("nothing new on the server\n"))
+
+                # "Activate" a tipmost bookmark.
+                bms = getattr(self.repo['tip'], 'bookmarks',
+                              lambda : None)()
+                if bms:
+                    bookmarks.setcurrent(self.repo, bms[0])
+
+        def remoteref(ref):
+            rn = remote_name or 'default'
+            return 'refs/remotes/' + rn + ref[10:]
+
+        modheads = [refs[k] for k in refs if k.startswith('refs/heads/')
+                    and not k.endswith('^{}')
+                    and refs[k] != oldrefs.get(remoteref(k))]
+
+        if not modheads:
+            self.ui.status(_("no changes found\n"))
 
         self.save_map()
 
+        return len(modheads)
+
     def export_commits(self):
         try:
             self.export_git_objects()
@@ -131,17 +209,20 @@ class GitHandler(object):
             return {}
 
         try:
-            client.send_pack(path, changed, None)
+            client.send_pack(path, changed, lambda have, want: [])
 
             changed_refs = [ref for ref, sha in new_refs.iteritems()
                             if sha != old_refs.get(ref)]
             new = [bin(self.map_hg_get(new_refs[ref])) for ref in changed_refs]
-            old = dict( (bin(self.map_hg_get(old_refs[r])), 1)
-                       for r in changed_refs if r in old_refs)
+            old = {}
+            for r in old_refs:
+                old_ref = self.map_hg_get(old_refs[r])
+                if old_ref:
+                    old[bin(old_ref)] = 1
 
             return old, new
-        except HangupException:
-            raise hgutil.Abort("the remote end hung up unexpectedly")
+        except (HangupException, GitProtocolError), e:
+            raise hgutil.Abort(_("git remote error: ") + str(e))
 
     def push(self, remote, revs, force):
         self.export_commits()
@@ -167,17 +248,39 @@ class GitHandler(object):
         if os.path.exists(mapfile):
             os.remove(mapfile)
 
+    # incoming support
+    def getremotechanges(self, remote, revs):
+        self.export_commits()
+        refs = self.fetch_pack(remote.path, revs)
+
+        # refs contains all remote refs. Prune to only those requested.
+        if revs:
+            reqrefs = {}
+            for rev in revs:
+                for n in ('refs/heads/' + rev, 'refs/tags/' + rev):
+                    if n in refs:
+                        reqrefs[n] = refs[n]
+        else:
+            reqrefs = refs
+
+        commits = [bin(c) for c in self.getnewgitcommits(reqrefs)[1]]
+
+        b = overlayrepo(self, commits, refs)
+
+        return (b, commits, lambda: None)
+
     ## CHANGESET CONVERSION METHODS
 
     def export_git_objects(self):
-        self.ui.status(_("importing Hg objects into Git\n"))
         self.init_if_missing()
 
         nodes = [self.repo.lookup(n) for n in self.repo]
         export = [node for node in nodes if not hex(node) in self._map_hg]
         total = len(export)
+        if total:
+            self.ui.status(_("exporting hg objects to git\n"))
         for i, rev in enumerate(export):
-            util.progress(self.ui, 'import', i, total=total)
+            util.progress(self.ui, 'exporting', i, total=total)
             ctx = self.repo.changectx(rev)
             state = ctx.extra().get('hg-git', None)
             if state == 'octopus':
@@ -185,7 +288,7 @@ class GitHandler(object):
                               "of octopus explosion\n" % ctx.rev())
                 continue
             self.export_hg_commit(rev)
-        util.progress(self.ui, 'import', None, total=total)
+        util.progress(self.ui, 'importing', None, total=total)
 
 
     # convert this commit into git objects
@@ -215,6 +318,10 @@ class GitHandler(object):
             # work around a timezone format change
             if int(timezone) % 60 != 0: #pragma: no cover
                 timezone = parse_timezone(timezone)
+                # Newer versions of Dulwich return a tuple here
+                if isinstance(timezone, tuple):
+                    timezone, neg_utc = timezone
+                    commit._commit_timezone_neg_utc = neg_utc
             else:
                 timezone = -int(timezone)
             commit.commit_timezone = timezone
@@ -244,22 +351,27 @@ class GitHandler(object):
         self.swap_out_encoding(oldenc)
         return commit.id
 
+    def get_valid_git_username_email(self, name):
+        return name.lstrip('< ').rstrip('> ')
+
     def get_git_author(self, ctx):
         # hg authors might not have emails
         author = ctx.user()
 
         # check for git author pattern compliance
-        regex = re.compile('^(.*?) \<(.*?)\>(.*)$')
+        regex = re.compile('^(.*?) ?\<(.*?)(?:\>(.*))?$')
         a = regex.match(author)
 
         if a:
             name = a.group(1)
             email = a.group(2)
-            if len(a.group(3)) > 0:
+            if a.group(3) != None and len(a.group(3)) != 0:
                 name += ' ext:(' + urllib.quote(a.group(3)) + ')'
-            author = name + ' <' + email + '>'
+            author = self.get_valid_git_username_email(name) + ' <' + self.get_valid_git_username_email(email) + '>'
+        elif '@' in author:
+            author = self.get_valid_git_username_email(author) + ' <' + self.get_valid_git_username_email(author) + '>'
         else:
-            author = author + ' <none@none>'
+            author = self.get_valid_git_username_email(author) + ' <none@none>'
 
         if 'author' in ctx.extra():
             author = "".join(apply_delta(author, ctx.extra()['author']))
@@ -344,8 +456,7 @@ class GitHandler(object):
 
             yield f, blobid, mode
 
-    def import_git_objects(self, remote_name=None, refs=None):
-        self.ui.status(_("importing Git objects into Hg\n"))
+    def getnewgitcommits(self, refs=None):
         self.init_if_missing()
 
         # import heads and fetched tags as remote references
@@ -364,7 +475,7 @@ class GitHandler(object):
                 if sha in self.git.object_store:
                     obj = self.git.get_object(sha)
                     while isinstance(obj, Tag):
-                        obj_type, sha = obj._get_object()
+                        obj_type, sha = obj.object
                         obj = self.git.get_object(sha)
                     if isinstance (obj, Commit) and sha not in seenheads:
                         seenheads.add(sha)
@@ -398,14 +509,34 @@ class GitHandler(object):
                 done.add(sha)
                 todo.pop()
 
-        commits = [commit for commit in commits if not commit in self._map_git]
+        return convert_list, [commit for commit in commits if not commit in self._map_git]
+
+    def import_git_objects(self, remote_name=None, refs=None):
+        convert_list, commits = self.getnewgitcommits(refs)
         # import each of the commits, oldest first
         total = len(commits)
+        if total:
+            self.ui.status(_("importing git objects into hg\n"))
+
         for i, csha in enumerate(commits):
-            util.progress(self.ui, 'import', i, total=total, unit='commits')
+            util.progress(self.ui, 'importing', i, total=total, unit='commits')
             commit = convert_list[csha]
             self.import_git_commit(commit)
-        util.progress(self.ui, 'import', None, total=total, unit='commits')
+        util.progress(self.ui, 'importing', None, total=total, unit='commits')
+
+        # Remove any dangling tag references.
+        for name, rev in self.repo.tags().items():
+            if not rev in self.repo:
+                if hasattr(self, 'tagscache') and self.tagscache and \
+                        'name' in self.tagscache:
+                    # Mercurial 1.4 and earlier.
+                    del self.repo.tagscache[name]
+                elif hasattr(self, '_tags') and self._tags and \
+                        'name' in self._tags:
+                    # Mercurial 1.5 and later.
+                    del self.repo._tags[name]
+                if name in self.repo._tagtypes:
+                    del self.repo._tagtypes[name]
 
     def import_git_commit(self, commit):
         self.ui.debug(_("importing: %s\n") % commit.id)
@@ -453,14 +584,38 @@ class GitHandler(object):
 
         oldenc = self.swap_out_encoding()
 
-        def getfilectx(repo, memctx, f):
-            delete, mode, sha = files[f]
-            if delete:
-                raise IOError
+        def findconvergedfiles(p1, p2):
+            # If any files have the same contents in both parents of a merge
+            # (and are therefore not reported as changed by Git) but are at
+            # different file revisions in Mercurial (because they arrived at
+            # those contents in different ways), we need to include them in
+            # the list of changed files so that Mercurial can join up their
+            # filelog histories (same as if the merge was done in Mercurial to
+            # begin with).
+            if p2 == nullid:
+                return []
+            manifest1 = self.repo.changectx(p1).manifest()
+            manifest2 = self.repo.changectx(p2).manifest()
+            return [path for path, node1 in manifest1.iteritems()
+                    if path not in files and manifest2.get(path, node1) != node1]
 
-            data = self.git[sha].data
-            copied_path = hg_renames.get(f)
-            e = self.convert_git_int_mode(mode)
+        def getfilectx(repo, memctx, f):
+            info = files.get(f)
+            if info != None:
+                # it's a file reported as modified from Git
+                delete, mode, sha = info
+                if delete:
+                    raise IOError
+
+                data = self.git[sha].data
+                copied_path = hg_renames.get(f)
+                e = self.convert_git_int_mode(mode)
+            else:
+                # it's a converged file
+                fc = context.filectx(self.repo, f, changeid=memctx.p1().rev())
+                data = fc.data()
+                e = fc.flags()
+                copied_path = fc.renamed()
 
             return context.memfilectx(f, data, 'l' in e, 'x' in e, copied_path)
 
@@ -471,8 +626,9 @@ class GitHandler(object):
         if len(gparents) > 1:
             # merge, possibly octopus
             def commit_octopus(p1, p2):
-                ctx = context.memctx(self.repo, (p1, p2), text, list(files), getfilectx,
-                                     author, date, {'hg-git': 'octopus'})
+                ctx = context.memctx(self.repo, (p1, p2), text,
+                                     list(files) + findconvergedfiles(p1, p2),
+                                     getfilectx, author, date, {'hg-git': 'octopus'})
                 return hex(self.repo.commitctx(ctx))
 
             octopus = len(gparents) > 2
@@ -521,8 +677,9 @@ class GitHandler(object):
         if not (repo_contains(p1) and repo_contains(p2)):
             raise hgutil.Abort(_('you appear to have run strip - '
                                  'please run hg git-cleanup'))
-        ctx = context.memctx(self.repo, (p1, p2), text, list(files), getfilectx,
-                             author, date, extra)
+        ctx = context.memctx(self.repo, (p1, p2), text,
+                             list(files) + findconvergedfiles(p1, p2),
+                             getfilectx, author, date, extra)
 
         node = self.repo.commitctx(ctx)
 
@@ -545,8 +702,8 @@ class GitHandler(object):
             self.ui.status(_("creating and sending data\n"))
             changed_refs = client.send_pack(path, changed, genpack)
             return changed_refs
-        except HangupException:
-            raise hgutil.Abort("the remote end hung up unexpectedly")
+        except (HangupException, GitProtocolError), e:
+            raise hgutil.Abort(_("git remote error: ") + str(e))
 
     def get_changed_refs(self, refs, revs, force):
         new_refs = refs.copy()
@@ -556,14 +713,26 @@ class GitHandler(object):
             del new_refs['capabilities^{}']
             if not self.local_heads():
                 tip = hex(self.repo.lookup('tip'))
-                bookmarks.bookmark(self.ui, self.repo, 'master', tip, force=True)
+                try:
+                    commands.bookmark(self.ui, self.repo, 'master', tip, force=True)
+                except NameError:
+                    bookmarks.bookmark(self.ui, self.repo, 'master', tip, force=True)
                 bookmarks.setcurrent(self.repo, 'master')
                 new_refs['refs/heads/master'] = self.map_git_get(tip)
 
         for rev in revs:
             ctx = self.repo[rev]
-            heads = [t for t in ctx.tags() if t in self.local_heads()]
-            tags = [t for t in ctx.tags() if t in self.tags]
+            if getattr(ctx, 'bookmarks', None):
+                labels = lambda c: ctx.tags() + [
+                                fltr for fltr, bm 
+                                in self._filter_for_bookmarks(ctx.bookmarks())
+                            ]
+            else:
+                labels = lambda c: ctx.tags()
+            prep = lambda itr: [i.replace(' ', '_') for i in itr]
+
+            heads = [t for t in prep(labels(ctx)) if t in self.local_heads()]
+            tags = [t for t in prep(labels(ctx)) if t in self.tags]
 
             if not (heads or tags):
                 raise hgutil.Abort("revision %s cannot be pushed since"
@@ -624,8 +793,13 @@ class GitHandler(object):
         def determine_wants(refs):
             if heads:
                 want = []
+                # contains pairs of ('refs/(heads|tags|...)/foo', 'foo')
+                # if ref is just '<foo>', then we get ('foo', 'foo')
+                stripped_refs = [
+                    (r, r[r.find('/', r.find('/')+1)+1:])
+                        for r in refs]
                 for h in heads:
-                    r = [ref for ref in refs if ref.endswith('/'+h)]
+                    r = [pair[0] for pair in stripped_refs if pair[1] == h]
                     if not r:
                         raise hgutil.Abort("ref %s not found on remote server" % h)
                     elif len(r) == 1:
@@ -634,15 +808,20 @@ class GitHandler(object):
                         raise hgutil.Abort("ambiguous reference %s: %r" % (h, r))
             else:
                 want = [sha for ref, sha in refs.iteritems()
-                        if not ref.endswith('^{}')]
+                        if not ref.endswith('^{}')
+                        and ( ref.startswith('refs/heads/') or ref.startswith('refs/tags/') ) ]
+            want = [x for x in want if x not in self.git]
             return want
         f, commit = self.git.object_store.add_pack()
         try:
             try:
-                return client.fetch_pack(path, determine_wants, graphwalker,
-                                         f.write, self.ui.status)
-            except HangupException:
-                raise hgutil.Abort("the remote end hung up unexpectedly")
+                progress = GitProgress(self.ui)
+                ret = client.fetch_pack(path, determine_wants, graphwalker,
+                                        f.write, progress.progress)
+                progress.flush()
+                return ret
+            except (HangupException, GitProtocolError), e:
+                raise hgutil.Abort(_("git remote error: ") + str(e))
         finally:
             commit()
 
@@ -654,21 +833,36 @@ class GitHandler(object):
         # Create a local Git branch name for each
         # Mercurial bookmark.
         for key in heads:
-            self.git.refs['refs/heads/' + key] = self.map_git_get(heads[key])
+            git_ref = self.map_git_get(heads[key])
+            if git_ref:
+                self.git.refs['refs/heads/' + key] = self.map_git_get(heads[key])
 
     def export_hg_tags(self):
         for tag, sha in self.repo.tags().iteritems():
             if self.repo.tagtype(tag) in ('global', 'git'):
+                tag = tag.replace(' ', '_')
                 self.git.refs['refs/tags/' + tag] = self.map_git_get(hex(sha))
                 self.tags[tag] = hex(sha)
 
+    def _filter_for_bookmarks(self, bms):
+        if not self.branch_bookmark_suffix:
+            return [(bm, bm) for bm in bms]
+        else:
+            def _filter_bm(bm):
+                if bm.endswith(self.branch_bookmark_suffix):
+                    return bm[0:-(len(self.branch_bookmark_suffix))]
+                else:
+                    return bm
+            return [(_filter_bm(bm), bm) for bm in bms]
+
     def local_heads(self):
         try:
             if getattr(bookmarks, 'parse', None):
                 bms = bookmarks.parse(self.repo)
             else:
                 bms = self.repo._bookmarks
-            return dict([(bm, hex(bms[bm])) for bm in bms])
+            return dict([(filtered_bm, hex(bms[bm])) for 
+                        filtered_bm, bm in self._filter_for_bookmarks(bms)])
         except AttributeError: #pragma: no cover
             return {}
 
@@ -694,7 +888,7 @@ class GitHandler(object):
                         sha = self.map_hg_get(refs[k])
                         self.tags[ref_name] = sha
                     elif isinstance (obj, Tag): # annotated
-                        (obj_type, obj_sha) = obj._get_object()
+                        (obj_type, obj_sha) = obj.object
                         obj = self.git.get_object(obj_sha)
                         if isinstance (obj, Commit):
                             sha = self.map_hg_get(obj_sha)
@@ -709,6 +903,7 @@ class GitHandler(object):
                 bms = bookmarks.parse(self.repo)
             else:
                 bms = self.repo._bookmarks
+
             heads = dict([(ref[11:],refs[ref]) for ref in refs
                           if ref.startswith('refs/heads/')])
 
@@ -726,6 +921,22 @@ class GitHandler(object):
                     if bm.ancestor(self.repo[hgsha]) == bm:
                         # fast forward
                         bms[head] = hgsha
+
+            # if there's a branch bookmark suffix,
+            # then add it on to all bookmark names
+            # that would otherwise conflict with a branch
+            # name
+            if self.branch_bookmark_suffix:
+                real_branch_names = self.repo.branchmap()
+                bms = dict(
+                    (
+                        bm_name + self.branch_bookmark_suffix 
+                            if bm_name in real_branch_names
+                        else bm_name,
+                        bms[bm_name]
+                    )
+                    for bm_name in bms
+                )
             if heads:
                 if oldbm:
                     bookmarks.write(self.repo, bms)
@@ -738,26 +949,35 @@ class GitHandler(object):
                          ' bookmarks enabled?\n'))
 
     def update_remote_branches(self, remote_name, refs):
-        def _set_hg_tag(head, sha):
-            # refs contains all the refs in the server, not just the ones
-            # we are pulling
-            if sha not in self.git.object_store:
-                return
-            hgsha = bin(self.map_hg_get(sha))
-            tag = '%s/%s' % (remote_name, head)
-            self.repo.tag(tag, hgsha, '', True, None, None)
-
+        tagfile = self.repo.join(os.path.join('git-remote-refs'))
+        tags = self.repo.gitrefs()
+        # since we re-write all refs for this remote each time, prune
+        # all entries matching this remote from our tags list now so
+        # that we avoid any stale refs hanging around forever
+        for t in list(tags):
+            if t.startswith(remote_name + '/'):
+                del tags[t]
+        tags = dict((k, hex(v)) for k, v in tags.iteritems())
+        store = self.git.object_store
         for ref_name, sha in refs.iteritems():
             if ref_name.startswith('refs/heads'):
+                if sha not in store:
+                    continue
+                hgsha = self.map_hg_get(sha)
                 head = ref_name[11:]
-                _set_hg_tag(head, sha)
-
+                tags['/'.join((remote_name, head))] = hgsha
+                # TODO(durin42): what is this doing?
                 new_ref = 'refs/remotes/%s/%s' % (remote_name, head)
                 self.git.refs[new_ref] = sha
             elif (ref_name.startswith('refs/tags')
                   and not ref_name.endswith('^{}')):
                 self.git.refs[ref_name] = sha
 
+        tf = open(tagfile, 'wb')
+        for tag, node in tags.iteritems():
+            tf.write('%s %s\n' % (node, tag))
+        tf.close()
+
 
     ## UTILITY FUNCTIONS
 
@@ -872,6 +1092,10 @@ class GitHandler(object):
             return string.decode('ascii', 'replace').encode('utf-8')
 
     def get_transport_and_path(self, uri):
+        # pass hg's ui.ssh config to dulwich
+        if not issubclass(client.get_ssh_vendor, _ssh.SSHVendor):
+            client.get_ssh_vendor = _ssh.generate_ssh_vendor(self.ui)
+
         for handler, transport in (("git://", client.TCPGitClient),
                                    ("git@", client.SSHGitClient),
                                    ("git+ssh://", client.SSHGitClient)):
@@ -889,11 +1113,32 @@ class GitHandler(object):
                 else:
                     hostpath_seper = '/'
 
+                port = None
                 host, path = hostpath.split(hostpath_seper, 1)
                 if hostpath_seper == '/':
                     transportpath = '/' + path
                 else:
-                    transportpath = path
-                return transport(host, thin_packs=False), transportpath
+                    # port number should be recognized
+                    m = re.match('^(?P<port>\d+)?(?P<path>.*)$', path)
+                    if m.group('port'):
+                        client.port = m.group('port')
+                        port = client.port
+                        transportpath = m.group('path')
+                    else:
+                        transportpath = path
+
+                return transport(host, thin_packs=False, port=port), transportpath
+
+        httpclient = getattr(client, 'HttpGitClient', None)
+
+        if uri.startswith('git+http://') or uri.startswith('git+https://'):
+            uri = uri[4:]
+
+        if uri.startswith('http://') or uri.startswith('https://'):
+            if not httpclient:
+                raise RepoError('git via HTTP requires dulwich 0.8.1 or later')
+            else:
+                return client.HttpGitClient(uri, thin_packs=False), uri
+
         # if its not git or git+ssh, try a local url..
         return client.SubprocessGitClient(thin_packs=False), uri
